@@ -19,6 +19,7 @@
  */
 
 #include "apr_strings.h"
+#include "apr_lib.h"
 #include "ap_config.h"
 #include "httpd.h"
 #include "http_config.h"
@@ -37,9 +38,14 @@ typedef enum {
     SLASH_UNSET
 } slash_cfg;
 
+#define REDIRECT_OFF   0
+#define REDIRECT_UNSET 1
+
 typedef struct dir_config_struct {
     apr_array_header_t *index_names;
     slash_cfg do_slash;
+    int redirect_index;
+    const char *dflt;
 } dir_config_rec;
 
 #define DIR_CMD_PERMS OR_INDEXES
@@ -47,11 +53,28 @@ typedef struct dir_config_struct {
 static const char *add_index(cmd_parms *cmd, void *dummy, const char *arg)
 {
     dir_config_rec *d = dummy;
+    const char *t, *w;
+    int count = 0;
 
     if (!d->index_names) {
         d->index_names = apr_array_make(cmd->pool, 2, sizeof(char *));
     }
-    *(const char **)apr_array_push(d->index_names) = arg;
+
+    t = arg;
+    while ((w = ap_getword_conf(cmd->pool, &t)) && w[0]) {
+        if (count == 0 && !strcasecmp(w, "disabled")) {
+            /* peek to see if "disabled" is first in a series of arguments */
+            const char *tt = t;
+            const char *ww = ap_getword_conf(cmd->temp_pool, &tt);
+            if (ww == NULL || !ww[0]) {
+               /* "disabled" is first, and alone */
+               break;
+            }
+        }
+        *(const char **)apr_array_push(d->index_names) = w;
+        count++;
+    }
+
     return NULL;
 }
 
@@ -62,13 +85,46 @@ static const char *configure_slash(cmd_parms *cmd, void *d_, int arg)
     d->do_slash = arg ? SLASH_ON : SLASH_OFF;
     return NULL;
 }
+static const char *configure_redirect(cmd_parms *cmd, void *d_, const char *arg1)
+{
+    dir_config_rec *d = d_;
+    int status;
 
+    if (!strcasecmp(arg1, "ON"))
+        status = HTTP_MOVED_TEMPORARILY;
+    else if (!strcasecmp(arg1, "OFF"))
+        status = REDIRECT_OFF;
+    else if (!strcasecmp(arg1, "permanent"))
+        status = HTTP_MOVED_PERMANENTLY;
+    else if (!strcasecmp(arg1, "temp"))
+        status = HTTP_MOVED_TEMPORARILY;
+    else if (!strcasecmp(arg1, "seeother"))
+        status = HTTP_SEE_OTHER;
+    else if (apr_isdigit(*arg1)) {
+        status = atoi(arg1);
+        if (!ap_is_HTTP_REDIRECT(status)) {
+            return "DirectoryIndexRedirect only accepts values between 300 and 399";
+        }
+    }
+    else {
+        return "DirectoryIndexRedirect ON|OFF|permanent|temp|seeother|3xx";
+    }
+
+    d->redirect_index = status;
+    return NULL;
+}
 static const command_rec dir_cmds[] =
 {
-    AP_INIT_ITERATE("DirectoryIndex", add_index, NULL, DIR_CMD_PERMS,
+    AP_INIT_TAKE1("FallbackResource", ap_set_string_slot,
+                  (void*)APR_OFFSETOF(dir_config_rec, dflt),
+                  DIR_CMD_PERMS, "Set a default handler"),
+    AP_INIT_RAW_ARGS("DirectoryIndex", add_index, NULL, DIR_CMD_PERMS,
                     "a list of file names"),
     AP_INIT_FLAG("DirectorySlash", configure_slash, NULL, DIR_CMD_PERMS,
                  "On or Off"),
+    AP_INIT_TAKE1("DirectoryIndexRedirect", configure_redirect,
+                   NULL, DIR_CMD_PERMS, "On, Off, or a 3xx status code."),
+
     {NULL}
 };
 
@@ -78,6 +134,7 @@ static void *create_dir_config(apr_pool_t *p, char *dummy)
 
     new->index_names = NULL;
     new->do_slash = SLASH_UNSET;
+    new->redirect_index = REDIRECT_UNSET;
     return (void *) new;
 }
 
@@ -90,9 +147,65 @@ static void *merge_dir_configs(apr_pool_t *p, void *basev, void *addv)
     new->index_names = add->index_names ? add->index_names : base->index_names;
     new->do_slash =
         (add->do_slash == SLASH_UNSET) ? base->do_slash : add->do_slash;
+    new->redirect_index=
+        (add->redirect_index == REDIRECT_UNSET) ? base->redirect_index : add->redirect_index;
+    new->dflt = add->dflt ? add->dflt : base->dflt;
     return new;
 }
 
+static int fixup_dflt(request_rec *r)
+{
+    dir_config_rec *d = ap_get_module_config(r->per_dir_config, &dir_module);
+    const char *name_ptr;
+    request_rec *rr;
+    int error_notfound = 0;
+
+    name_ptr = d->dflt;
+    if (name_ptr == NULL) {
+        return DECLINED;
+    }
+    /* XXX: if FallbackResource points to something that doesn't exist,
+     * this may recurse until it hits the limit for internal redirects
+     * before returning an Internal Server Error.
+     */
+
+    /* The logic of this function is basically cloned and simplified
+     * from fixup_dir below.  See the comments there.
+     */
+    if (r->args != NULL) {
+        name_ptr = apr_pstrcat(r->pool, name_ptr, "?", r->args, NULL);
+    }
+    rr = ap_sub_req_lookup_uri(name_ptr, r, r->output_filters);
+    if (rr->status == HTTP_OK
+        && (   (rr->handler && !strcmp(rr->handler, "proxy-server"))
+            || rr->finfo.filetype == APR_REG)) {
+        ap_internal_fast_redirect(rr, r);
+        return OK;
+    }
+    else if (ap_is_HTTP_REDIRECT(rr->status)) {
+
+        apr_pool_join(r->pool, rr->pool);
+        r->notes = apr_table_overlay(r->pool, r->notes, rr->notes);
+        r->headers_out = apr_table_overlay(r->pool, r->headers_out,
+                                           rr->headers_out);
+        r->err_headers_out = apr_table_overlay(r->pool, r->err_headers_out,
+                                               rr->err_headers_out);
+        error_notfound = rr->status;
+    }
+    else if (rr->status && rr->status != HTTP_NOT_FOUND
+             && rr->status != HTTP_OK) {
+        error_notfound = rr->status;
+    }
+
+    ap_destroy_sub_req(rr);
+    if (error_notfound) {
+        return error_notfound;
+    }
+
+    /* nothing for us to do, pass on through */
+    return DECLINED;
+}
+
 static int fixup_dir(request_rec *r)
 {
     dir_config_rec *d;
@@ -101,11 +214,6 @@ static int fixup_dir(request_rec *r)
     int num_names;
     int error_notfound = 0;
 
-    /* only handle requests against directories */
-    if (r->finfo.filetype != APR_DIR) {
-        return DECLINED;
-    }
-    
     /* In case mod_mime wasn't present, and no handler was assigned. */
     if (!r->handler) {
         r->handler = DIR_MAGIC_TYPE;
@@ -120,7 +228,7 @@ static int fixup_dir(request_rec *r)
                                                &dir_module);
 
     /* Redirect requests that are not '/' terminated */
-    if (r->uri[0] == '\0' || r->uri[strlen(r->uri) - 1] != '/') 
+    if (r->uri[0] == '\0' || r->uri[strlen(r->uri) - 1] != '/')
     {
         char *ifile;
 
@@ -129,8 +237,8 @@ static int fixup_dir(request_rec *r)
         }
 
         /* Only redirect non-get requests if we have no note to warn
-         * that this browser cannot handle redirs on non-GET requests 
-         * (such as Microsoft's WebFolders). 
+         * that this browser cannot handle redirs on non-GET requests
+         * (such as Microsoft's WebFolders).
          */
         if ((r->method_number != M_GET)
                 && apr_table_get(r->subprocess_env, "redirect-carefully")) {
@@ -151,10 +259,6 @@ static int fixup_dir(request_rec *r)
         return HTTP_MOVED_PERMANENTLY;
     }
 
-    if (strcmp(r->handler, DIR_MAGIC_TYPE)) {
-        return DECLINED;
-    }
-
     if (d->index_names) {
         names_ptr = (char **)d->index_names->elts;
         num_names = d->index_names->nelts;
@@ -180,16 +284,31 @@ static int fixup_dir(request_rec *r)
             name_ptr = apr_pstrcat(r->pool, name_ptr, "?", r->args, NULL);
         }
 
-        rr = ap_sub_req_lookup_uri(name_ptr, r, NULL);
+        rr = ap_sub_req_lookup_uri(name_ptr, r, r->output_filters);
 
-        /* XXX: (filetype == APR_REG) - we can't use a non-file index??? */
-        if (   rr->status == HTTP_OK
+        /* The sub request lookup is very liberal, and the core map_to_storage
+         * handler will almost always result in HTTP_OK as /foo/index.html
+         * may be /foo with PATH_INFO="/index.html", or even / with
+         * PATH_INFO="/foo/index.html". To get around this we insist that the
+         * the index be a regular filetype.
+         *
+         * Another reason is that the core handler also makes the assumption
+         * that if r->finfo is still NULL by the time it gets called, the
+         * file does not exist.
+         */
+        if (rr->status == HTTP_OK
             && (   (rr->handler && !strcmp(rr->handler, "proxy-server"))
                 || rr->finfo.filetype == APR_REG)) {
+
+            if (ap_is_HTTP_REDIRECT(d->redirect_index)) {
+                apr_table_setn(r->headers_out, "Location", ap_construct_url(r->pool, rr->uri, r));
+                return d->redirect_index;
+            }
+
             ap_internal_fast_redirect(rr, r);
             return OK;
         }
-            
+
         /* If the request returned a redirect, propagate it to the client */
 
         if (ap_is_HTTP_REDIRECT(rr->status)
@@ -227,16 +346,35 @@ static int fixup_dir(request_rec *r)
         return error_notfound;
     }
 
+    /* record what we tried, mostly for the benefit of mod_autoindex */
+    apr_table_set(r->notes, "dir-index-names",
+                  d->index_names ?
+                  apr_array_pstrcat(r->pool, d->index_names, ','):
+                  AP_DEFAULT_INDEX);
+
     /* nothing for us to do, pass on through */
     return DECLINED;
 }
 
+static int dir_fixups(request_rec *r)
+{
+    if (r->finfo.filetype == APR_DIR) {
+        /* serve up a directory */
+        return fixup_dir(r);
+    }
+    else if ((r->finfo.filetype == APR_NOFILE) && (r->handler == NULL)) {
+        /* No handler and nothing in the filesystem - use fallback */
+        return fixup_dflt(r);
+    }
+    return DECLINED;
+}
+
 static void register_hooks(apr_pool_t *p)
 {
-    ap_hook_fixups(fixup_dir,NULL,NULL,APR_HOOK_LAST);
+    ap_hook_fixups(dir_fixups,NULL,NULL,APR_HOOK_LAST);
 }
 
-module AP_MODULE_DECLARE_DATA dir_module = {
+AP_DECLARE_MODULE(dir) = {
     STANDARD20_MODULE_STUFF,
     create_dir_config,          /* create per-directory config structure */
     merge_dir_configs,          /* merge per-directory config structures */
